""" python syntax highlighting tool

"""

import click
import sys
import yaml

from importlib.resources import read_text

from . import VERSION
from .render import OnscreenRender
from .color import Color

from . import resources
import io

# Fun comment here
#


@click.command()
@click.argument("input-file", type=click.File(mode="rb"), default=sys.stdin)
@click.argument("output-file", type=click.Path(allow_dash=True), default=sys.stdout)
@click.option("-l", "--start-line", default=0, help="line to begin displaying")
@click.option("-n", "--line-count", default=10, help="number of lines to display")
@click.option("-p", "--preview", is_flag=True, default=False)
@click.option("-t", "--transparent", is_flag=True, default=False)
@click.option("-s", "--style-name", type=click.STRING, default="default")
@click.option("--list-styles", is_flag=True, default=False)
@click.option("-d", "--debug", is_flag=True, hidden=True, default=False)
@click.version_option(VERSION)
def pyliter_cli(
    input_file,
    output_file,
    start_line,
    line_count,
    preview,
    transparent,
    style_name,
    list_styles,
    debug,
):
    """Python syntax highlighting

    Performs Python syntax highlighting on code found in INPUT_FILE
    and writes color annotated text in PNG format to OUTPUT_FILE.
    """

    style = yaml.safe_load(read_text(resources, "default_style.yaml"))

    for category, attributes in style.items():
        for color_key in ["color", "background_color", "underline"]:
            try:
                color_spec = attributes[color_key]
                color = Color.from_any(color_spec)
                if color_key == "background_color" and transparent:
                    color.alpha.value = 0
                attributes[color_key] = color.rgba
            except KeyError:
                pass

    if input_file == sys.stdin:
        input_file = io.BytesIO(sys.stdin.buffer.read())

    render = OnscreenRender(input_file, start_line, line_count, style, output_file)

    render.run(preview)



""" pyglet application
"""
import pyglet
from .document import PythonDocument
from .color import Color

# Comment goes here
#         some more


class OnscreenRender(pyglet.window.Window):
    def __init__(
        self,
        fileobj,
        start_line,
        line_count,
        style_book: dict = None,
        output_path: str = None,
        debug: bool = False,
    ):

        self.debug = debug
        self.output_path = output_path or "screenshot.png"

        self.document = PythonDocument(fileobj, start_line, line_count, style_book)

        width, height = self.document.dimensions

        config = pyglet.gl.Config(
            depth=32,
            sample_buffers=1,
            samples=4,
            double_buffer=True,
            depth_buffer=0,
            stereo=0,
            alpha_size=8,
        )

        try:
            caption = fileobj.name
        except AttributeError:
            caption = "stdin"

        super().__init__(
            width=width + (self.margin * 2),
            height=height + (self.margin * 2),
            config=config,
            visible=True,
            caption=fileobj.name,
        )

        pyglet.gl.glEnable(pyglet.gl.GL_BLEND)
        pyglet.gl.glBlendFunc(pyglet.gl.GL_SRC_ALPHA, pyglet.gl.GL_ONE_MINUS_SRC_ALPHA)
        pyglet.gl.glClearColor(*Color(*self.document.background_color).rgba_f)

        self.layout.x = self.margin
        self.layout.y = self.margin

        if self.debug:
            self.batch.add(
                8,
                pyglet.gl.GL_LINES,
                self.background,
                ("v2i", self.text_bounds),
                ("c3B", (255, 0, 0) * 8),
            )
        self.active = True

    @property
    def margin(self):
        try:
            return self._margin
        except AttributeError:
            pass
        self._margin = self.document.font.height * 2
        return self._margin

    @property
    def batch(self):
        try:
            return self._batch
        except AttributeError:
            pass
        self._batch = pyglet.graphics.Batch()
        return self._batch

    @property
    def foreground(self):
        try:
            return self._foreground
        except AttributeError:
            pass
        self._foreground = pyglet.graphics.OrderedGroup(0)
        return self._foreground

    @property
    def background(self):
        try:
            return self._background
        except AttributeError:
            pass
        self._background = pyglet.graphics.OrderedGroup(1)
        return self._background

    @property
    def layout(self):
        try:
            return self._layout
        except AttributeError:
            pass

        self._layout = pyglet.text.layout.TextLayout(
            self.document,
            self.width - (self.margin * 2),
            self.height - (self.margin * 2),
            multiline=True,
            wrap_lines=False,
            batch=self.batch,
            group=self.foreground,
        )
        return self._layout

    def on_draw(self):
        self.clear()
        if self.debug:
            pyglet.gl.glColor4f(1, 0, 0, 1)
            pyglet.gl.glLineWidth(3)
        self.batch.draw()
        pyglet.image.get_buffer_manager().get_color_buffer().save(self.output_path)

    def on_show(self):

        self.on_draw()
        self.active = False

    @property
    def text_bounds(self):
        try:
            return self._text_bounds
        except AttributeError:
            pass

        x = self.layout.x
        y = self.layout.y
        w = x + self.layout.width
        h = x + self.layout.height + 1

        self._text_bounds = (x, y, x, h, x, h, w, h, w, h, w, y, w, y, x, y)
        return self._text_bounds

    def run(self, preview: bool = False):

        if preview:
            pyglet.app.run()
        else:
            while self.active:
                self.dispatch_events()



import pyglet

from .token import ExtendedToken as PythonTokens
from .token import DOCSTRING


class PythonDocument(pyglet.text.document.FormattedDocument):
    def __init__(
        self, fileobj, start: int = 0, line_count: int = -1, style_book: dict = None
    ):

        self.style_book = style_book or {
            "DEFAULT": {
                "font_name": "Source Code Pro",
                "font_size": 18,
                "color": (255, 255, 255, 255),
                "background_color": (0, 0, 0, 255),
                "line_numbers": False,
            }
        }
        self.first = start
        self.last = start + line_count

        self.tokens, self.raw_text = PythonTokens.from_file(fileobj, start, line_count)

        super().__init__(self.raw_text.translate(str.maketrans({"{": "{{", "}": "}}"})))

        self.line_count = len(self.text.splitlines()) + 1
        # FormattedDocument does some messing about with newlines in the source
        # text which results in the loss of the terminal newline. This in turn
        # causes an off-by-one bug when counting the number of lines in the text.

        self.set_style(
            start=0, end=len(self.text), attributes=self.style_book["DEFAULT"]
        )

        self._apply_token_styles()

        if self.style_book["DEFAULT"].get("line_numbers", False):
            self._insert_line_numbers(start)

    def _insert_line_numbers(self, start):
        """
        """
        p = 0
        line_num = start
        attributes = {
            "color": (0xA0, 0xA0, 0xA0, 0xFF),
            "underline": None,
            "bold": None,
            "italic": None,
        }
        while True:
            try:
                self.insert_text(p, f"{line_num:>4d} ", attributes=attributes)
                p = self.text.index("\n", p) + 1
                line_num += 1
            except ValueError:
                break

    def _apply_token_styles(self, tokens: list = None) -> None:
        """
        """
        tokens = tokens or self.tokens

        p = 0

        for token in tokens:
            if not token.category in self.style_book:
                continue

            try:
                if token.exact_type == DOCSTRING:
                    p = self.text.index('"""', p)
                    l = self.text.index('"""', p + 1) + 3
                else:
                    p = self.text.index(token.string, p)
                    l = len(token.string)
            except ValueError:
                print("F", p, token.string)
                print(repr(token))
                continue

            self.set_style(p, p + l, self.style_book[token.category])
            p += l

    @property
    def background_color(self) -> tuple:
        """
        """
        return self.style_book["DEFAULT"]["background_color"]

    @property
    def color(self) -> tuple:
        """
        """
        return self.style_book["DEFAULT"]["color"]

    @property
    def dimensions(self) -> tuple:
        """
        """
        try:
            return self._dimensions
        except AttributeError:
            pass
        self._dimensions = (
            self.max_column * self.font.size,
            self.line_count * self.font.height,
        )

        return self._dimensions

    @property
    def max_column(self) -> int:
        """
        """
        try:
            return self._max_column
        except AttributeError:
            pass
        self._max_column = max(len(l.strip()) for l in self.raw_text.splitlines())
        return self._max_column

    @property
    def font(self):
        """
        """
        try:
            return self._font
        except AttributeError:
            pass

        self._font = self.get_font(0)

        self._font.height = self._font.ascent - self._font.descent

        return self._font
